/*******************************************************************************
 * Copyright 2008, 2012 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package kdo.ebn;

/**
 * Class representing an Effect within a extended behavior network.
 */
public class Effect extends Proposition
{
	/** for degree in which the effect will come true */
	private Proposition degree;

	/** the probability in which the effect will come true */
	private double probability;

	/**
	 * Default constructor. Sets probability to value 1.0 and degree to null
	 */
	public Effect()
	{
		probability = 1.0;
		degree = null;
	}

	/**
	 * Constructor which sets the internal attributes of this effect object.
	 *
	 * @param parent the containing network node of this proposition
	 * @param perceptionNode perception used to calculate effect value
	 * @param isNegated negation value for the effect
	 * @param probability the probability with which an effect comes true
	 * @param degree to which an effect comes true
	 */
	public Effect(NetworkNode parent, PerceptionNode perceptionNode, boolean isNegated, double probability,
			Proposition degree)
	{
		super(parent, perceptionNode, isNegated);
		this.probability = probability;
		this.degree = degree;
	}

	/**
	 * Sets the internal attributes of this effect object.
	 *
	 * @param perceptionNode perception used to calculate effect value
	 * @param isNegated negation value for the effect
	 * @param probability the probability with which an effect comes true
	 * @param degree the degree to which an effect comes true
	 */
	public void setEffect(PerceptionNode perceptionNode, boolean isNegated, double probability, Proposition degree)
	{
		setProposition(perceptionNode, isNegated);
		this.probability = probability;
		this.degree = degree;
	}

	/**
	 * @return the probability with which an effect comes true
	 */
	public double getProbability()
	{
		return probability;
	}

	/**
	 * @return the degree to which an effect comes true
	 */
	double getDegree()
	{
		return (degree == null) ? 1.0 : degree.getTruthValue();
	}

	/**
	 * @return degree proposition
	 */
	public Proposition getDegreeProposition()
	{
		return degree;
	}

	/**
	 * @return the influence an effect has on activation spreading as product of
	 *         degree and probability
	 */
	public double getInfluence()
	{
		return getProbability() * getDegree();
	}
}
