/*******************************************************************************
 * Copyright 2008, 2012 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package kdo.ebn.xml.converter;

import java.util.Iterator;
import java.util.Map;

import kdo.ebn.Competence;
import kdo.ebn.ExtendedBehaviorNetwork;
import kdo.ebn.Goal;
import kdo.ebn.IEBNAction;
import kdo.ebn.IEBNPerception;
import kdo.ebn.NetworkConfigurationException;
import kdo.ebn.NetworkParams;
import kdo.ebn.PerceptionNode;
import kdo.ebn.Resource;

import com.thoughtworks.xstream.converters.Converter;
import com.thoughtworks.xstream.converters.MarshallingContext;
import com.thoughtworks.xstream.converters.UnmarshallingContext;
import com.thoughtworks.xstream.io.HierarchicalStreamReader;
import com.thoughtworks.xstream.io.HierarchicalStreamWriter;

/**
 * XStream converter to convert ExtendedBehaviorNetwork objects
 * @author Thomas Rinklin
 */
public class ExtendedBehaviorNetworkConverter implements Converter
{
	@SuppressWarnings("unused")
	private final Map<String, IEBNPerception> beliefs;

	private final Map<String, IEBNAction> behaviors;

	/**
	 * constructor with the maps of beliefs and behaviors
	 * @param beliefs map of beliefs (and resources)
	 * @param behaviors map of behaviors
	 */
	public ExtendedBehaviorNetworkConverter(Map<String, IEBNPerception> beliefs, Map<String, IEBNAction> behaviors)
	{
		this.behaviors = behaviors;
		this.beliefs = beliefs;
	}

	@Override
	public void marshal(Object value, HierarchicalStreamWriter writer, MarshallingContext context)
	{
		ExtendedBehaviorNetwork ebn = (ExtendedBehaviorNetwork) value;

		// NetworkParams
		writer.startNode("NetworkParams");
		context.convertAnother(ebn.getNetworkParams());
		writer.endNode();

		// Perceptions
		writer.startNode("Perceptions");
		Iterator<PerceptionNode> percIter = ebn.getPerceptions();
		while (percIter.hasNext()) {
			PerceptionNode perceptionNode = percIter.next();
			writer.startNode("PerceptionNode");
			context.convertAnother(perceptionNode);
			writer.endNode();
		}
		writer.endNode();

		// Ressources
		writer.startNode("Resources");
		Iterator<Resource> resIter = ebn.getResources();
		while (resIter.hasNext()) {
			Resource resource = resIter.next();
			writer.startNode("Resource");
			context.convertAnother(resource);
			writer.endNode();
		}
		writer.endNode();

		// Goals
		writer.startNode("Goals");
		Iterator<Goal> goalIter = ebn.getGoals();

		while (goalIter.hasNext()) {
			Goal goal = goalIter.next();
			writer.startNode("Goal");
			context.convertAnother(goal);
			writer.endNode();
		}
		writer.endNode();

		// Note, that Competences MUST be added AFTER the goals
		// Competences
		writer.startNode("Competences");
		Iterator<Competence> compIter = ebn.getCompetenceModules();
		while (compIter.hasNext()) {
			Competence competence = compIter.next();

			writer.startNode("Competence");
			context.convertAnother(competence);
			writer.endNode();
		}
		writer.endNode();
	}

	@Override
	public Object unmarshal(HierarchicalStreamReader reader, UnmarshallingContext context)
	{
		ExtendedBehaviorNetwork ebn = new ExtendedBehaviorNetwork("", behaviors);

		while (reader.hasMoreChildren()) {
			reader.moveDown();

			String nodeName = reader.getNodeName();

			if ("NetworkParams".equals(nodeName)) {
				NetworkParams newNetParams = (NetworkParams) context.convertAnother(ebn, NetworkParams.class);
				ebn.getNetworkParams().setNetworkParams(newNetParams);

			} else if ("Perceptions".equals(nodeName)) {
				while (reader.hasMoreChildren()) {
					reader.moveDown();
					if ("PerceptionNode".equalsIgnoreCase(reader.getNodeName())) {
						PerceptionNode perc = (PerceptionNode) context.convertAnother(ebn, PerceptionNode.class);
						ebn.addPerception(perc);
					}
					reader.moveUp();
				}

			} else if ("Resources".equals(nodeName)) {
				while (reader.hasMoreChildren()) {
					reader.moveDown();
					if ("Resource".equals(reader.getNodeName())) {
						Resource res = (Resource) context.convertAnother(ebn, Resource.class);
						ebn.addResource(res);
					}
					reader.moveUp();
				}

			} else if ("Goals".equals(nodeName)) {
				int goalCount = 0;
				while (reader.hasMoreChildren()) {
					reader.moveDown();
					if ("Goal".equals(reader.getNodeName())) {
						Goal goal = (Goal) context.convertAnother(ebn, Goal.class);
						ebn.addGoal(goal);
						goalCount++;
					}
					reader.moveUp();
				}
				context.put("GoalCount", goalCount);

			} else if ("Competences".equals(nodeName)) {
				while (reader.hasMoreChildren()) {
					reader.moveDown();
					if ("Competence".equals(reader.getNodeName())) {
						Competence comp = (Competence) context.convertAnother(ebn, Competence.class);
						try {
							ebn.addCompetenceModule(comp);
						} catch (NetworkConfigurationException e) {
							e.printStackTrace();
						}
					}
					reader.moveUp();
				}
			}
			reader.moveUp();
		}

		return ebn;
	}

	@SuppressWarnings("rawtypes")
	@Override
	public boolean canConvert(Class arg0)
	{
		return arg0.equals(ExtendedBehaviorNetwork.class);
	}
}
