/*
 * Copyright (c) 2008 Klaus Dorer
 * Hochschule Offenburg
 */
package kdo.neuralNetwork.feedforward;

import kdo.neuralNetwork.output.IOutputFunction;
import kdo.neuralNetwork.output.IdentityOutputFunction;
import kdo.neuralNetwork.transfer.SigmoidTransferFunction;
import kdo.util.IRandomSource;

/**
 * Represents a multi-layer feed-forward network using backpropagation
 */
public class Backpropagation extends FeedForwardNetwork
{
	/** */
	private static final long serialVersionUID = 1L;

	/** the layers of weights of this network */
	private final Layer[] layers;

	/**
	 * @param noOfNodes array with the number of neurons (nodes) in each layer
	 * @param initialLearnRate the initial value of the learnrate alpha
	 * @param rand the source for random numbers
	 */
	public Backpropagation(int[] noOfNodes, float initialLearnRate, IRandomSource rand)
	{
		this(noOfNodes, initialLearnRate, rand, null, new IdentityOutputFunction());
	}

	/**
	 * @param noOfNodes array with the number of neurons (nodes) in each layer
	 * @param initialLearnRate the initial value of the learnrate alpha
	 * @param rand the source for random numbers
	 * @param weights the weights to use for the network, null if random
	 */
	public Backpropagation(int[] noOfNodes, float initialLearnRate, IRandomSource rand, float[][][] weights,
			IOutputFunction outputFunction)
	{
		super(initialLearnRate, rand, new SigmoidTransferFunction(), outputFunction);
		layers = new Layer[noOfNodes.length - 1];
		int i;
		for (i = 0; i < noOfNodes.length - 2; i++) {
			layers[i] = new BPGHiddenLayer(noOfNodes[i], noOfNodes[i + 1], transfer, rand);
		}
		layers[i] = new BPGOutputLayer(noOfNodes[i], noOfNodes[i + 1], transfer, rand);

		if (weights != null) {
			int j = 0;
			for (Layer current : layers) {
				current.initializeWeights(weights[j++]);
			}
		} else {
			initializeWeights(0.0f, 1.0f);
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see kdo.neuralNetwork.INeuralNetwork#initializeWeights(float, float)
	 */
	public void initializeWeights(float standardValue, float delta)
	{
		for (Layer current : layers) {
			current.initializeWeights(standardValue, delta);
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see kdo.neuralNetwork.INeuralNetwork#recall(float[])
	 */
	public float[] recall(float[] input)
	{
		float[] currentInput = input;
		for (Layer current : layers) {
			// System.out.println("Layer Input: " + Arrays.toString(currentInput));
			currentInput = current.recall(currentInput);
		}
		// System.out.println("Output: " + Arrays.toString(currentInput));
		return outputFunction.output(currentInput);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see kdo.neuralNetwork.INeuralNetwork#train(float[], float[])
	 */
	public float train(float[] input, float[] expectedOutput)
	{
		float error = recallError(input, expectedOutput);
		float[] delta = expectedOutput;

		for (int i = layers.length - 1; i >= 0; i--) {
			Layer currentLayer = layers[i];
			float[] newDelta = currentLayer.calculateDelta(delta);
			delta = currentLayer.propagateDelta(newDelta);
			currentLayer.adjustWeights(newDelta, learnrate);
		}

		// float[] output = layers[layers.length - 1].outputLayer;
		// System.out.println("Input: " + Arrays.toString(input) + " Output: "
		// + Arrays.toString(output) + " Expected: "
		// + Arrays.toString(expectedOutput));

		return error;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		StringBuilder builder = new StringBuilder();
		for (Layer layer : layers) {
			builder.append(layer.toString());
		}
		return builder.toString();
	}
}
