/*
 * Copyright (c) 2008 Klaus Dorer
 * Hochschule Offenburg
 */
package kdo.neuralNetwork.feedforward;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.List;

import kdo.neuralNetwork.INeuralNetwork;
import kdo.neuralNetwork.data.InOutPattern;
import kdo.neuralNetwork.output.IOutputFunction;
import kdo.neuralNetwork.transfer.ITransferFunction;
import kdo.util.IRandomSource;

/**
 * Base class of feed forward networks
 */
public abstract class FeedForwardNetwork implements INeuralNetwork, Serializable
{
	/**	 */
	private static final long serialVersionUID = 1L;

	/** the initial learnrate */
	protected float learnrate;

	/** the source for random to be used */
	protected IRandomSource random;

	/** the transfer function used to transfer activation */
	protected ITransferFunction transfer;

	/** the output function used to output activation */
	protected final IOutputFunction outputFunction;

	/**
	 *
	 * @param initialLearnRate the initial value of the learnrate alpha
	 * @param rand the source for random numbers
	 */
	public FeedForwardNetwork(float initialLearnRate, IRandomSource rand, ITransferFunction transferFunction,
			IOutputFunction outputFunction)
	{
		random = rand;
		learnrate = initialLearnRate;
		transfer = transferFunction;
		this.outputFunction = outputFunction;
	}

	/**
	 * Activation function
	 * @param value the value for which to calculate the function
	 * @return the activation as transfered by transfer function
	 */
	protected float transfer(float value)
	{
		return transfer.transfer(value);
	}

	@Override
	public float recallError(float[] input, float[] output)
	{
		float[] out = recall(input);
		float result = 0.0f;
		for (int i = 0; i < out.length; i++) {
			float delta = out[i] - output[i];
			result += delta * delta;
		}
		// System.out.println("input: " + Arrays.toString(input) + "output: "
		// + Arrays.toString(out) + " goaloutput: " + Arrays.toString(output));
		return result;
	}

	public double recallAllRMS(List<InOutPattern> patterns)
	{
		double error = 0.0f;
		for (InOutPattern current : patterns) {
			float singleError = recallError(current.getInput(), current.getOutput());
			error += singleError;
			// System.out.println("error: " + singleError);
		}
		error = Math.sqrt(error / patterns.size());
		// System.out.println("\nError: " + error);
		return error;
	}

	public double recallAllPercent(List<InOutPattern> patterns)
	{
		int missclassified = 0;
		for (InOutPattern current : patterns) {
			float singleError = recallError(current.getInput(), current.getOutput());
			if (singleError > 0.01) {
				missclassified++;
			}
		}
		double error = (double) missclassified / patterns.size();
		// System.out.println("\nError: " + error);
		return error;
	}

	public double trainAll(List<InOutPattern> patterns, int maxLoops)
	{
		double error = 1.0f;
		int loops = 0;
		for (; loops < maxLoops && error > 0.0001f; loops++) {
			error = 0.0f;
			for (InOutPattern current : patterns) {
				float singleError = train(current.getInput(), current.getOutput());
				error += singleError;
				// System.out.println("error: " + singleError);
			}
			printError(error);
		}
		error = Math.sqrt(error / patterns.size());
		// System.out.println("\nError: " + error + " Loops: " + loops);
		return error;
	}

	protected void printError(double error)
	{
		System.out.println("Error; " + error);
	}

	public static INeuralNetwork loadNetwork(String filename)
	{
		try {
			ObjectInputStream in = new ObjectInputStream(new FileInputStream(filename));
			INeuralNetwork network = (INeuralNetwork) in.readObject();
			in.close();
			return network;

		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return null;
	}

	public static void saveNetwork(INeuralNetwork network, String filename)
	{
		try {
			ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(filename));
			out.writeObject(network);
			out.flush();
			out.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
