/*
 * Copyright (c) 2008 Klaus Dorer
 * Hochschule Offenburg
 */
package kdo.neuralNetwork.feedforward;

import java.io.Serializable;
import java.util.Arrays;

import kdo.neuralNetwork.transfer.ITransferFunction;
import kdo.util.IRandomSource;

/**
 * Represents a layer of nodes and their incoming weights in a neural network
 */
public abstract class Layer implements Serializable
{
	/** */
	private static final long serialVersionUID = 1L;

	/** the weights to the output layer */
	protected float[][] weight;

	/** the output of each node */
	protected final float[] outputLayer;

	/** the input coming from previous layer */
	protected final float[] inputLayer;

	/** the transfer function used for the nodes in this layer */
	protected final ITransferFunction transferFunction;

	/** the source for random to be used */
	private final IRandomSource random;

	/**
	 *
	 * @param noOfInputs the number of input neurons (including bias)
	 * @param noOfOutputs the number of output neurons
	 * @param transfer the transfer function to use
	 * @param rand the source for random numbers
	 */
	public Layer(int noOfInputs, int noOfOutputs, ITransferFunction transfer, IRandomSource rand)
	{
		if (noOfInputs < 1) {
			throw new IllegalArgumentException("Need at least one input but had: " + noOfInputs);
		}
		if (noOfOutputs < 1) {
			throw new IllegalArgumentException("Need at least one output but had: " + noOfOutputs);
		}
		random = rand;
		transferFunction = transfer;
		weight = new float[noOfInputs][noOfOutputs];
		outputLayer = new float[noOfOutputs];
		inputLayer = new float[noOfInputs];
		initializeWeights(0.0f, 1.0f);
	}

	/**
	 * Initializes all weights of this layer with values that are uniformly
	 * distributed between standardValue - delta (including) and standardValue +
	 * delta (not including)
	 * @param standardValue mean value for the weights
	 * @param delta maximal delta to the standardValue
	 */
	public void initializeWeights(float standardValue, float delta)
	{
		for (float[] row : weight) {
			for (int j = 0; j < row.length; j++) {
				row[j] = random.nextFloat() * 2.0f * delta + standardValue - delta;
			}
		}
	}

	/**
	 * Initializes all weights of this layer with values passed
	 * @param weights array of weights
	 * @param delta maximal delta to the standardValue
	 */
	public void initializeWeights(float[][] weights)
	{
		checkInputLength(weights.length);
		checkOutputLength(weights[0].length);
		this.weight = weights;
	}

	/**
	 * Calculates the output of this layer for the input passed
	 * @param input the input for which to calculate the output. The length has
	 *        to fit to the number of inputs.
	 * @return the output of this layer
	 */
	public float[] recall(float[] input)
	{
		checkInputLength(input.length);

		// store input
		System.arraycopy(input, 0, inputLayer, 0, input.length);

		// calculate output
		int outputs = weight[0].length;
		float[] localOutput = new float[outputs];
		for (int i = 0; i < outputs; i++) {
			for (int j = 0; j < weight.length; j++) {
				localOutput[i] += weight[j][i] * input[j];
			}
			localOutput[i] = transferFunction.transfer(localOutput[i]);
			outputLayer[i] = localOutput[i];
		}
		return localOutput;
	}

	/**
	 * Adjusts the weights of this layer using the learn strategy
	 * @param deltaOutput the delta of the output
	 * @param learnrate the current learnrate
	 * @return the error in this pattern as the mean of all delta squares of
	 *         output and expectedOutput
	 */
	public float adjustWeights(float[] deltaOutput, float learnrate)
	{
		checkOutputLength(deltaOutput.length);

		// adjust weights
		float error = 0.0f;
		for (int i = 0; i < deltaOutput.length; i++) {
			for (int j = 0; j < weight.length; j++) {
				float delta = deltaOutput[i];
				error += delta * delta;
				weight[j][i] += learnrate * inputLayer[j] * delta;
			}
		}

		// System.out.println("Weights: " + Arrays.deepToString(weight));

		error /= deltaOutput.length;
		return error;
	}

	/**
	 * @param inputLength the number of neurons in the input
	 */
	private void checkInputLength(int inputLength)
	{
		if (inputLength != weight.length) {
			throw new IllegalArgumentException(
					"Input size " + inputLength + " does not match network input size " + weight.length);
		}
	}

	/**
	 * @param outputLength the number of neurons in the output
	 */
	private void checkOutputLength(int outputLength)
	{
		if (outputLength != weight[0].length) {
			throw new IllegalArgumentException(
					"Output size " + outputLength + " does not match network output size " + weight[0].length);
		}
	}

	/**
	 * @param expectedOutput
	 * @return
	 */
	public abstract float[] calculateDelta(float[] expectedOutput);

	/**
	 * @param deltaNextLayer
	 * @return
	 */
	public float[] propagateDelta(float[] deltaNextLayer)
	{
		float[] result = new float[weight.length];
		for (int i = 0; i < weight.length; i++) {
			float sum = 0.0f;
			for (int k = 0; k < deltaNextLayer.length; k++) {
				sum += deltaNextLayer[k] * weight[i][k];
			}
			result[i] = sum;
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		return Arrays.deepToString(weight) + "\n";
	}
}
