/**
 *
 */
package kdo.search;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import kdo.search.impl.SearchParameters;
import kdo.search.runs.averageOutStrategies.IAverageOutStrategy;
import kdo.search.runs.averageOutStrategies.IDomainVisitor;
import kdo.search.runs.averageOutStrategies.ParallelAverageOutStrategy;
import kdo.search.runs.averageOutStrategies.SerialAverageOutStrategy;
import kdo.search.runs.averageOutStrategies.ThreadedAverageOutStrategy;
import kdo.search.strategy.ILocalSearchStrategy;
import kdo.search.strategy.local.BasicVariableNeighborhoodSearch;
import kdo.search.strategy.local.FullSearch;
import kdo.search.strategy.local.HillClimbing;
import kdo.search.strategy.local.HillClimbingRandom;
import kdo.search.strategy.local.HillClimbingRandomRestart;
import kdo.search.strategy.local.HillClimbingStateCache;
import kdo.search.strategy.local.RandomWalk;
import kdo.search.strategy.local.TabuSearch;
import kdo.search.strategy.local.UtilityMeasurement;
import kdo.search.strategy.local.annealing.HyperbolicTemperatureReduction;
import kdo.search.strategy.local.annealing.ITemperatureReductionStrategy;
import kdo.search.strategy.local.annealing.LinearTemperatureReduction;
import kdo.search.strategy.local.annealing.SigmoidalTemperatureReduction;
import kdo.search.strategy.local.annealing.SimulatedAnnealing;
import kdo.search.strategy.local.cmaes.impl.CMAESOptimizerAdapter;
import kdo.search.strategy.local.threaded.TabuSearchMultiThreaded;

/**
 * @author kdorer
 */
public class LearningConfigurationHelper
{
	private static final String MEASUREMENT = "Measurement";

	public static final String GENETIC = "Genetic";

	private static final String FULL_SEARCH = "FullSearch";

	private static final String CMAES_SEARCH = "CMAES";

	// local search strategies
	private static final String VARIABLE_NEIGHBORHOOD_SEARCH = "VNS";

	private static final String TABU_SEARCH = "TabuSearch";

	private static final String TABU_SEARCH_THREADED = "TabuSearchThreaded";

	private static final String SIMULATED_ANNEALING = "SimulatedAnnealing";

	private static final String RANDOM_WALK = "RandomWalk";

	private static final String HILL_CLIMBING_RANDOM_OPERATOR = "HCRandomOperator";

	private static final String HILL_CLIMBING_RANDOM_RESTART = "HCRandomRestart";

	private static final String HILL_CLIMBING = "HillClimbing";

	private static final String HILL_CLIMBING_STATE_CACHE = "HillClimbingStateCache";

	// avg out strategies
	public static final String SERIAL = "serial";

	private static final String PARALLEL = "parallel";

	private static final String THREADED = "threaded";

	/**
	 * A read-only list of all available optimization strategies.
	 */
	public static final List<String> STRATEGIES;

	static
	{
		List<String> strategies = new ArrayList<>();
		strategies.addAll(Arrays.asList(MEASUREMENT, GENETIC, FULL_SEARCH, CMAES_SEARCH, VARIABLE_NEIGHBORHOOD_SEARCH,
				TABU_SEARCH, TABU_SEARCH_THREADED, SIMULATED_ANNEALING, RANDOM_WALK, HILL_CLIMBING_RANDOM_OPERATOR,
				HILL_CLIMBING_RANDOM_RESTART, HILL_CLIMBING_STATE_CACHE, HILL_CLIMBING));
		STRATEGIES = Collections.unmodifiableList(strategies);
	}

	/**
	 * A read-only list of all available avg strategies.
	 */
	public static final List<String> AVG_STRATEGIES;

	static
	{
		List<String> avgStrategies = new ArrayList<>();
		avgStrategies.addAll(Arrays.asList(SERIAL, PARALLEL, THREADED));
		AVG_STRATEGIES = Collections.unmodifiableList(avgStrategies);
	}

	/**
	 *
	 * @param avgStrategy strategy name
	 * @param avgOutRuns number of runs
	 * @param threads the number of threads for parallel strategy
	 * @return an average out strategy for
	 */
	public static IAverageOutStrategy getAverageOutStrategy(
			String avgStrategy, int avgOutRuns, int threads, boolean verbose, IDomainVisitor visitor)
	{
		switch (avgStrategy) {
		case SERIAL:
			return new SerialAverageOutStrategy(avgOutRuns, verbose, visitor);
		case PARALLEL:
			return new ParallelAverageOutStrategy(avgOutRuns, threads, verbose, visitor);
		case THREADED:
			return new ThreadedAverageOutStrategy(avgOutRuns, verbose, visitor);
		}
		return null;
	}

	/**
	 * Returns a local search method to be used. Genetic search is treated
	 * separately, although in principle it is also a LocalSearch and properly
	 * inheriting.
	 * @param params local search parameters
	 * @param threads number of parallel threads
	 * @return the local search strategy to be used
	 */
	public static ILocalSearchStrategy getLocalSearchStrategy(SearchParameters params, int threads)
	{
		String strategyString = params.getStrategy();
		long maxRuntime = params.getMaxRuntime();
		long maxIterations = (long) params.getStrategyParameter("MaxIterations", 10000000);

		switch (strategyString) {
		case HILL_CLIMBING:
			return new HillClimbing(maxRuntime);
		case HILL_CLIMBING_STATE_CACHE:
			int cacheSize = (int) params.getStrategyParameter("CacheSize", 5);
			return new HillClimbingStateCache(maxRuntime, maxIterations, cacheSize);
		case HILL_CLIMBING_RANDOM_RESTART:
			int restarts = (int) params.getStrategyParameter("Restarts", 100);
			return new HillClimbingRandomRestart(restarts, maxRuntime);
		case HILL_CLIMBING_RANDOM_OPERATOR:
			return new HillClimbingRandom(maxIterations, maxRuntime);
		case RANDOM_WALK:
			int randomSteps = (int) params.getStrategyParameter("RandomSteps", 30);
			return new RandomWalk(randomSteps, maxIterations, maxRuntime);
		case SIMULATED_ANNEALING:
			int temperatureStrategy = (int) params.getStrategyParameter("TemperatureStrategy", 0);
			double startTemperature = params.getStrategyParameter("StartTemperature", 0.5);
			double endTemperature = params.getStrategyParameter("EndTemperature", 0.0);
			ITemperatureReductionStrategy reductionStrategy = null;
			switch (temperatureStrategy) {
			case 1:
				reductionStrategy = new SigmoidalTemperatureReduction(startTemperature, endTemperature);
				break;
			case 2:
				reductionStrategy = new HyperbolicTemperatureReduction(startTemperature, endTemperature);
				break;
			default:
				reductionStrategy = new LinearTemperatureReduction(startTemperature, endTemperature);
			}
			return new SimulatedAnnealing(reductionStrategy, maxIterations, maxRuntime);
		case TABU_SEARCH:
			int tabulistSize = (int) params.getStrategyParameter("TabulistSize", 100);
			return new TabuSearch(tabulistSize, maxIterations, maxRuntime);
		case TABU_SEARCH_THREADED:
			tabulistSize = (int) params.getStrategyParameter("TabulistSize", 100);
			return new TabuSearchMultiThreaded(tabulistSize, maxIterations, maxRuntime, threads);
		case VARIABLE_NEIGHBORHOOD_SEARCH:
			// do basic variable neighborhood search
			int neighborhoodRadius = (int) params.getStrategyParameter("NeighborhoodRadius", 175);
			return new BasicVariableNeighborhoodSearch(new HillClimbing(-1), neighborhoodRadius, maxRuntime);
		case MEASUREMENT:
			return new UtilityMeasurement();
		case FULL_SEARCH:
			return new FullSearch();
		case CMAES_SEARCH:
			int populationSize = (int) params.getStrategyParameter("PopulationSize", 50);
			int generations = (int) params.getStrategyParameter("Generations", 70);
			double stopFitness = params.getStrategyParameter("StopFitness", 100);
			double sigmaSpread = params.getStrategyParameter("Sigma", 0.1);
			return CMAESOptimizerAdapter.getInstance(populationSize, generations, stopFitness, sigmaSpread);
		}
		return null;
	}
}
